@@ -68,10 +68,10 @@ enum json_ast_obj_type_t
 
 struct json_ast_obj_t
 {
-    enum json_ast_obj_type_t type;
+    enum json_ast_obj_type_t    type;
     union {
-        struct json_ast_node_t*  node;
-        struct json_text_pos_t   mark;
+        struct json_ast_node_t* node;
+        struct json_text_pos_t  mark;
     };
 };
 
@@ -96,8 +96,9 @@ struct json_ast_obj_t
 
 #define JSON_AST_STACK_SIZE()       STACK_SIZE(&ast->obj_stack)
 #define JSON_AST_STACK_POP()        STACK_POP(&ast->obj_stack)
-#define JSON_AST_STACK_TOP_N(n)     STACK_TOP_N(&ast->obj_stack, n)
 #define JSON_AST_STACK_POP_N(n)     STACK_POP_N(&ast->obj_stack, n)
+#define JSON_AST_STACK_TOP_N_REF(n) STACK_TOP_N_REF(&ast->obj_stack, n)
+#define JSON_AST_STACK_TOP_REF()    STACK_TOP_REF(&ast->obj_stack)
 
 static size_t json_ast_stack_find_first_mark(
     struct json_ast_stack_t* stack)
@@ -351,34 +352,40 @@ static bool json_ast_object_start(struct json_ast_t* ast)
     return true;
 }
 
-#define ENSURE_JSON_AST_OBJ_IS(o, n)                       \
-    do {                                                   \
-        ENSURE((o).type == json_ast_obj_ ## n ## _type,    \
-            "invalid obj stack: element is not '" #n "'"); \
-    } while (0)
+#define JSON_AST_OBJ_IS(o, n)                    \
+    (                                            \
+        (o)->type == json_ast_obj_ ## n ## _type \
+    )
+
+#define ENSURE_JSON_AST_OBJ_IS(o, n)                      \
+        ENSURE(JSON_AST_OBJ_IS(o, n),                     \
+            "invalid obj stack: element is not '" #n "'")
 
 #define ENSURE_JSON_AST_NODE_OBJ_IS(o, n)                       \
     do {                                                        \
-        ENSURE((o).node->type == json_ast_ ## n ## _node_type,  \
+        ENSURE_JSON_AST_OBJ_IS(o, node);                        \
+        ENSURE((o)->node->type == json_ast_ ## n ## _node_type, \
             "invalid obj stack: node is not '" #n "' but '%s'", \
-            json_ast_node_type_get_name((o).node->type));       \
+            json_ast_node_type_get_name((o)->node->type));      \
     } while (0)
 
 static bool json_ast_object_end(struct json_ast_t* ast)
 {
-    struct json_ast_obj_t o;
-    struct json_text_pos_t w;
+    struct json_ast_object_arg_t* p = NULL;
     struct json_ast_node_t* n;
-    struct json_ast_object_arg_t *p = NULL, *q;
-    size_t i, k, s = 0;
+    struct json_text_pos_t w;
+    struct json_ast_obj_t* o;
+    size_t i, s = 0;
 
     i = json_ast_stack_find_first_mark(
             &ast->obj_stack);
-    o = JSON_AST_STACK_TOP_N(i);
-    ENSURE_JSON_AST_OBJ_IS(o, mark);
-    w = o.mark;
+    o = JSON_AST_STACK_TOP_N_REF(i);
+    ASSERT(JSON_AST_OBJ_IS(o, mark));
+    w = o->mark;
 
     if (i > 0) {
+        struct json_ast_obj_t *b, *q;
+
         ENSURE((i % 2) == 0,
             "invalid obj stack: number of elements "
             "until mark is odd");
@@ -387,23 +394,25 @@ static bool json_ast_object_end(struct json_ast_t* ast)
         p = JSON_AST_ALLOCATE_ARRAY(
                 struct json_ast_object_arg_t, s);
 
-        for (k = i - 1, q = p; ; k -= 2, q ++) {
-            o = JSON_AST_STACK_TOP_N(k);
-            ENSURE_JSON_AST_OBJ_IS(o, node);
-            ENSURE_JSON_AST_NODE_OBJ_IS(o, string);
+        for (b = JSON_AST_STACK_TOP_REF(),
+             q = b - SIZE_AS_INT(i - 1);
+             q < b;
+             q ++,
+             p ++) {
+            ENSURE_JSON_AST_NODE_OBJ_IS(q, string);
 
-            q->key = &o.node->string;
+            p->key = &q->node->string;
             ASSERT(
-                JSON_AST_TO_NODE(q->key, string) ==
-                o.node); 
+                JSON_AST_TO_NODE(p->key, string) ==
+                q->node); 
 
-            o = JSON_AST_STACK_TOP_N(k - 1);
-            ENSURE_JSON_AST_OBJ_IS(o, node);
+            q ++;
+            ENSURE_JSON_AST_OBJ_IS(q, node);
 
-            q->val = o.node;
-
-            if (k == 1) break;
+            p->val = q->node;
         }
+
+        p -= s;
     }
     JSON_AST_STACK_POP_N(i);
 
@@ -425,22 +434,26 @@ static bool json_ast_array_start(struct json_ast_t* ast)
 static void json_ast_array_end_pos(struct json_ast_t* ast,
     size_t pos, const struct json_text_pos_t* where, bool mark)
 {
-    struct json_ast_obj_t o;
-    struct json_ast_node_t *n, **p = NULL, **q;
-    size_t k, s = pos;
+    struct json_ast_node_t *n, **p = NULL;
+    size_t s = pos;
 
     if (pos > 0) {
+        struct json_ast_obj_t *b, *q;
+
         p = JSON_AST_ALLOCATE_ARRAY(
                 struct json_ast_node_t*, pos);
 
-        for (k = pos - 1, q = p; ; k --) {
-            o = JSON_AST_STACK_TOP_N(k);
-            ENSURE_JSON_AST_OBJ_IS(o, node);
+        for (b = JSON_AST_STACK_TOP_REF(),
+             q = b - SIZE_AS_INT(pos - 1);
+             q <= b;
+             q ++,
+             p ++) {
+            ENSURE_JSON_AST_OBJ_IS(q, node);
 
-            *q ++ = o.node;
-
-            if (k == 0) break;
+            *p = q->node;
         }
+
+        p -= pos;
     }
     if (!mark) {
         ASSERT_SIZE_DEC_NO_OVERFLOW(s);
@@ -454,21 +467,19 @@ static void json_ast_array_end_pos(struct json_ast_t* ast,
 
 static bool json_ast_array_end(struct json_ast_t* ast)
 {
-    struct json_ast_obj_t o;
-    struct json_text_pos_t w;
+    struct json_ast_obj_t* o;
     size_t i;
 
     i = json_ast_stack_find_first_mark(
             &ast->obj_stack);
-    o = JSON_AST_STACK_TOP_N(i);
-    ENSURE_JSON_AST_OBJ_IS(o, mark);
-    w = o.mark;
+    o = JSON_AST_STACK_TOP_N_REF(i);
+    ASSERT(JSON_AST_OBJ_IS(o, mark));
 
-    json_ast_array_end_pos(ast, i, &w, true);
+    json_ast_array_end_pos(ast, i, &o->mark, true);
     return true;
 }
 
-static struct json_handler_t handler = {
+static const struct json_handler_t handler = {
     .null_func         = (json_null_func_t)         json_ast_null,
     .boolean_func      = (json_boolean_func_t)      json_ast_boolean,
     .number_func       = (json_number_func_t)       json_ast_number,
@@ -500,12 +511,14 @@ struct json_ast_t* json_ast_create(
         handler.object_key_func == 
         handler.string_func);
 
-    ast->json_obj = json_obj_create(&handler, ast, &sizes->obj);
+    ast->json_obj = 
+        json_obj_create(&handler, ast, &sizes->obj);
 
     pool_alloc_init(
         &ast->pool_alloc, sizes->pool_size);
     json_ast_stack_init(
-        &ast->obj_stack, sizes->obj.stack_max, sizes->obj.stack_init);
+        &ast->obj_stack, sizes->obj.stack_max,
+        sizes->obj.stack_init);
 
     return ast;
 }
@@ -576,7 +589,7 @@ enum json_parse_status_t json_ast_parse_done(
     ASSERT(n == 1);
 
     o = JSON_AST_STACK_POP();
-    ENSURE_JSON_AST_OBJ_IS(o, node);
+    ENSURE_JSON_AST_OBJ_IS(&o, node);
 
     ast->root = o.node;
     ast->done = true;
