@@ -1,4 +1,4 @@
-// Copyright (C) 2016, 2017, 2018  Stefan Vargyas
+// Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas
 // 
 // This file is part of Json-Type.
 // 
@@ -44,9 +44,6 @@ struct json_ast_node_t
     struct json_text_pos_t pos;
 };
 
-struct json_text_pos_t json_get_current_tok_pos(
-    struct json_obj_t* obj);
-
 #define JSON_AST_TYPEOF_IS_NODE_(q, p) \
     TYPEOF_IS(p, q struct json_ast_node_t*)
 
