@@ -1,4 +1,4 @@
-// Copyright (C) 2016, 2017, 2018  Stefan Vargyas
+// Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas
 // 
 // This file is part of Json-Type.
 // 
@@ -69,5 +69,14 @@ JSON_API void mem_buf_shift_fit(
         *(buf->ptr + buf->len ++) = ch;     \
     } while (0)
 
+#define MEM_BUF_REMOVE(buf, n)              \
+    do {                                    \
+        STATIC(MEM_BUF_TYPEOF_IS_BUF(buf)); \
+        STATIC(TYPEOF_IS_SIZET(n));         \
+        ASSERT(buf->len <= buf->size);      \
+        ASSERT(buf->len >= (n));            \
+        buf->len -= (n);                    \
+    } while (0)
+
 #endif/*__MEM_BUF_H*/
 
