@@ -1,4 +1,4 @@
-// Copyright (C) 2016, 2017, 2018  Stefan Vargyas
+// Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas
 // 
 // This file is part of Json-Type.
 // 
@@ -115,7 +115,7 @@ size_t pretty_print_char(FILE* file, char ch, size_t flags)
         fputc('\'', file);
 
     n = fprintf(file, pretty_print_fmt(ch, flags), (uchar_t) ch);
-    r = INT_AS_SIZE(n);
+    r = VERIFY_INT_AS_SIZE(n);
 
     if (q) {
         fputc('\'', file);
@@ -140,7 +140,7 @@ size_t pretty_print_string(
         int n;
 
         n = fprintf(file, pretty_print_fmt(c, flags), c);
-        s = INT_AS_SIZE(n);
+        s = VERIFY_INT_AS_SIZE(n);
 
         SIZE_ADD_EQ(r, s);
     }
@@ -153,6 +153,56 @@ size_t pretty_print_string(
     return r;
 }
 
+size_t pretty_print_strings(
+    FILE* file, const char* const* strs, size_t n_strs,
+    const char* name, size_t width, size_t flags)
+{
+    size_t i, r;
+
+    if (n_strs == 0) {
+        size_t l = strlen(name);
+        int c;
+
+        ASSERT_SIZE_INC_NO_OVERFLOW(l);
+
+        c = fprintf(file,
+                "%s:%*s-\n", name, width > l + 1
+                ? SIZE_AS_INT(width - (l + 1)) : 0, "");
+
+        return VERIFY_INT_AS_SIZE(c);
+    }
+
+    for (r = 0,
+         i = 0;
+         i < n_strs;
+         i ++, strs ++) {
+        size_t n;
+        int c;
+
+        c = fprintf(file, "%s[%zu]:", name, i);
+        n = VERIFY_INT_AS_SIZE(c);
+        SIZE_ADD_EQ(r, n);
+
+        if (width > n) {
+            c = fprintf(file, "%*s",
+                    SIZE_AS_INT(width - n), "");
+            n = VERIFY_INT_AS_SIZE(c);
+            SIZE_ADD_EQ(r, n);
+        }
+
+        n = pretty_print_string(file,
+                PTR_UCHAR_CAST_CONST(*strs),
+                strlen(*strs), flags);
+        SIZE_ADD_EQ(r, n);
+
+        c = fputc('\n', file);
+        VERIFY(c >= 0);
+
+        SIZE_PRE_INC(r);
+    }
+    return r;
+}
+
 static void pretty_print_repr_base(
     FILE* file, const uchar_t* str, size_t len,
     bool quote_non_print)
