@@ -1,4 +1,4 @@
-// Copyright (C) 2016, 2017, 2018  Stefan Vargyas
+// Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas
 // 
 // This file is part of Json-Type.
 // 
@@ -23,18 +23,17 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
-#include <errno.h>
 #include <stdio.h>
 #include <ctype.h>
 
 #include "common.h"
 #include "char-traits.h"
 #include "ptr-traits.h"
+#include "su-size.h"
 
 #include "test-common.h"
 
 #define ISASCII CHAR_IS_ASCII
-#define ISDIGIT CHAR_IS_DIGIT
 
 const char stdin_name[] = "<stdin>";
 
@@ -67,24 +66,6 @@ static void usage(void)
     fprintf(stdout, help, program);
 }
 
-struct su_size_t su_size(size_t sz)
-{
-    struct su_size_t r;
-
-    r.sz = sz;
-    r.su = "";
-    if (r.sz  % MB(1) == 0) {
-        r.sz /= MB(1);
-        r.su = "M";
-    }
-    else
-    if (r.sz  % KB(1) == 0) {
-        r.sz /= KB(1);
-        r.su = "K";
-    }
-    return r;
-}
-
 void options_invalid_opt_arg(const char* opt_name, const char* opt_arg)
 {
     error("invalid argument for '%s' option: '%s'", opt_name, opt_arg);
@@ -120,98 +101,10 @@ static void invalid_opt_ch(char opt_name)
     error("invalid command line option '-%c'", opt_name);
 }
 
-#if SIZE_MAX < ULONG_MAX
-static size_t strtosz(
-    const char* ptr, char** end, int base)
-{
-    unsigned long r;
-
-    errno = 0;
-    r = strtoul(ptr, end, base);
-
-    if (errno == 0 && r > SIZE_MAX)
-        errno = ERANGE;
-
-    return r;
-}
-#define STR_TO_SIZE strtosz
-#elif SIZE_MAX == ULONG_MAX
-#define STR_TO_SIZE strtoul
-#elif SIZE_MAX == UULONG_MAX
-#define STR_TO_SIZE strtoull
-#else
-#error unexpected SIZE_MAX > UULONG_MAX
-#endif
-
-size_t parse_num(
-    const char* p, const char** q, int b)
-{
-    if (!ISDIGIT(*p)) {
-        *q = p;
-        errno = EINVAL;
-        return 0;
-    }
-    errno = 0;
-    return STR_TO_SIZE(p, (char**) q, b);
-}
-
-size_t options_parse_size_optarg(
-    const char* opt_name, const char* opt_arg,
-    size_t min, size_t max)
-{
-    const char *p, *q = NULL;
-    size_t n, v, d;
-
-    if (!(n = strlen(opt_arg)))
-        options_invalid_opt_arg(opt_name, opt_arg);
-    v = parse_num(p = opt_arg, &q, 10);
-    d = PTR_OFFSET(q, p, n);
-    if (errno || (d != n))
-        options_invalid_opt_arg(opt_name, opt_arg);
-    if ((min > 0 && v < min) || (max > 0 && v > max))
-        options_illegal_opt_arg(opt_name, opt_arg);
-    return v;
-}
-
-size_t options_parse_su_size_optarg(
-    const char* opt_name, const char* opt_arg,
-    size_t min, size_t max)
-{
-    const char *p, *q = NULL;
-    size_t n, v, d;
-
-    if (!(n = strlen(opt_arg)))
-        options_invalid_opt_arg(opt_name, opt_arg);
-    v = parse_num(p = opt_arg, &q, 10);
-    d = PTR_OFFSET(q, p, n);
-    if (errno ||
-        (d == 0) ||
-        (d < n - 1) ||
-        (d == n - 1 && *q != 'k' && *q != 'K' &&
-            *q != 'm' && *q != 'M'))
-        options_invalid_opt_arg(opt_name, opt_arg);
-    switch (*q) {
-    case '\0':
-        break;
-    case 'k':
-    case 'K':
-        STATIC(KB(1) <= SIZE_MAX);
-        if (!SIZE_MUL_NO_OVERFLOW(v, KB(1)))
-            options_illegal_opt_arg(opt_name, opt_arg);
-        v *= KB(1);
-        break;
-    case 'm':
-    case 'M':
-        STATIC(MB(1) <= SIZE_MAX);
-        if (!SIZE_MUL_NO_OVERFLOW(v, MB(1)))
-            options_illegal_opt_arg(opt_name, opt_arg);
-        v *= MB(1);
-        break;
-    }
-    if ((min > 0 && v < min) || (max > 0 && v > max))
-        options_illegal_opt_arg(opt_name, opt_arg);
-    return v;
-}
+#define SU_SIZE_OPT_NAME options
+#define SU_SIZE_OPT_NEED_PARSE_SIZE
+#define SU_SIZE_OPT_NEED_PARSE_SIZE_SU
+#include "su-size-opt-impl.h"
 
 void options_parse_args(void* opts,
     const struct options_funcs_t* funcs,
