@@ -1,4 +1,4 @@
-// Copyright (C) 2016, 2017, 2018  Stefan Vargyas
+// Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas
 // 
 // This file is part of Json-Type.
 // 
@@ -29,7 +29,9 @@
 
 #include "common.h"
 #include "ptr-traits.h"
+#include "pretty-print.h"
 #include "bit-set.h"
+#include "su-size.h"
 
 #include "test-common.h"
 
@@ -127,8 +129,6 @@ static void options_dump(const struct options_t* opts)
         CASE(object),
         CASE(dict),
     };
-    char* const* p;
-    size_t i;
 
 #define NAME(x)  OPTIONS_NAME(x)
 #define NNUL(x)  OPTIONS_NNUL(x)
@@ -156,8 +156,9 @@ static void options_dump(const struct options_t* opts)
         opts->div,
         opts->argc);
 
-    for (i = 0, p = opts->argv; i < opts->argc; i ++, p ++)
-        fprintf(stdout, "argv[%zu]:     %s\n", i, *p);
+    pretty_print_strings(stdout,
+        PTR_CONST_PTR_CAST(opts->argv, char),
+        opts->argc, "argv", 13, 0);
 }
 
 enum {
@@ -661,19 +662,12 @@ static bool gen_expr_parse_alpha_key(
 static bool gen_expr_parse_num_key(
     const char* str, size_t* res)
 {
-    const char* p = NULL;
-    size_t r;
-
     if (str[0] == '0' &&
         str[1] != '\0')
         return false;
 
-    r = parse_num(str, &p, 10);
-    if (errno || *p)
-        return false;
-
-    *res = r;
-    return true;
+    return su_size_parse(
+        str, 0, 0, res, NULL);
 }
 
 enum gen_expr_type_t
