@@ -1,4 +1,4 @@
-// Copyright (C) 2016, 2017, 2018  Stefan Vargyas
+// Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas
 // 
 // This file is part of Json-Type.
 // 
@@ -31,7 +31,9 @@
 
 #if defined(TRIE_NEED_LOOKUP_SYM) || \
     defined(TRIE_NEED_INSERT_SYM) || \
-    defined(TRIE_NEED_NODE_HAS_NULL_SIB)
+    defined(TRIE_NEED_NODE_HAS_NULL_SIB) || \
+    defined(TRIE_NEED_GEN_CODE_GET_ATTR) || \
+    defined(TRIE_NEED_GEN_CODE)
 #ifndef TRIE_NULL_SYM
 #error  TRIE_NULL_SYM is not defined
 #endif
@@ -51,7 +53,8 @@
 #error  TRIE_SYM_CMP is not defined
 #endif
 
-#ifdef  TRIE_NEED_PRINT
+#if defined(TRIE_NEED_PRINT_NODE) || \
+    defined(TRIE_NEED_PRINT)
 
 #ifndef TRIE_SYM_PRINT
 #error  TRIE_SYM_PRINT is not defined
@@ -61,7 +64,8 @@
 #error  TRIE_VAL_PRINT is not defined
 #endif
 
-#endif // TRIE_NEED_PRINT
+#endif // TRIE_NEED_PRINT_NODE ||
+       // TRIE_NEED_PRINT
 
 #ifndef TRIE_KEY_TYPE
 #error  TRIE_KEY_TYPE is not defined
@@ -123,6 +127,22 @@
 
 #endif // TRIE_NEED_GEN_DEF
 
+#ifdef  TRIE_NEED_GEN_CODE
+
+#ifndef TRIE_SYM_GEN_CODE
+#error  TRIE_SYM_GEN_CODE is not defined
+#endif
+
+#ifndef TRIE_VAL_GEN_CODE
+#error  TRIE_VAL_GEN_CODE is not defined
+#endif
+
+#ifndef TRIE_ERR_GEN_CODE
+#error  TRIE_ERR_GEN_CODE is not defined
+#endif
+
+#endif // TRIE_NEED_GEN_CODE
+
 #ifndef TRIE_NEED_IMPL_ONLY 
 
 #ifndef TRIE_ALLOC_OBJ_TYPE
@@ -141,6 +161,7 @@
 
 #define TRIE_TYPE            TRIE_TYPE_(TRIE_NAME)
 #define TRIE_NODE_TYPE       TRIE_NODE_TYPE_(TRIE_NAME)
+#define TRIE_ATTR_TYPE       TRIE_MAKE_NAME(TRIE_NAME, node_attr_t)
 
 #define TRIE_ALLOC_FUNC_TYPE TRIE_MAKE_NAME(TRIE_NAME, alloc_func_t)
 #define TRIE_INIT            TRIE_MAKE_NAME(TRIE_NAME, init)
@@ -166,7 +187,9 @@
 
 #define TRIE_LOOKUP_SYM_NODE TRIE_MAKE_NAME(TRIE_NAME, lookup_sym_node)
 #define TRIE_LOOKUP_SYM      TRIE_MAKE_NAME(TRIE_NAME, lookup_sym)
+#define TRIE_LOOKUP_SYM_AT   TRIE_MAKE_NAME(TRIE_NAME, lookup_sym_at)
 #define TRIE_LOOKUP_KEY      TRIE_MAKE_NAME(TRIE_NAME, lookup_key)
+#define TRIE_LOOKUP_KEY_AT   TRIE_MAKE_NAME(TRIE_NAME, lookup_key_at)
 #define TRIE_INSERT_SYM      TRIE_MAKE_NAME(TRIE_NAME, insert_sym)
 #define TRIE_INSERT_KEY      TRIE_MAKE_NAME(TRIE_NAME, insert_key)
 
@@ -274,6 +297,18 @@
 #define TRIE_LVL_ITERATOR_DEREF  TRIE_MAKE_NAME(TRIE_NAME, lvl_iterator_deref)
 #define TRIE_LVL_ITERATOR_INC    TRIE_MAKE_NAME(TRIE_NAME, lvl_iterator_inc)
 
+#define TRIE_NODE_GEN_ATTRS      TRIE_MAKE_NAME(TRIE_NAME, node_gen_attrs)
+
+#define TRIE_NODE_GEN_CODE_GET_ATTR TRIE_MAKE_NAME(TRIE_NAME, node_gen_code_get_attr)
+#define TRIE_GEN_CODE_GET_ATTR   TRIE_MAKE_NAME(TRIE_NAME, gen_code_get_attr)
+#define TRIE_NODE_GEN_CODE       TRIE_MAKE_NAME(TRIE_NAME, node_gen_code)
+#define TRIE_GEN_CODE            TRIE_MAKE_NAME(TRIE_NAME, gen_code)
+
+#define TRIE_PATH_STACK_TYPE     TRIE_MAKE_NAME(TRIE_NAME, path_stack_t)
+#define TRIE_PATH_STACK_INIT     TRIE_MAKE_NAME(TRIE_NAME, path_stack_init)
+#define TRIE_PATH_STACK_DONE     TRIE_MAKE_NAME(TRIE_NAME, path_stack_done)
+#define TRIE_PATH_STACK_NAME     TRIE_MAKE_NAME(TRIE_NAME, path)
+
 // stev: the trie implementation below is due to the work of
 // Bentley and Sedgewick; see the web page and the papers
 // listed below:
@@ -292,8 +327,22 @@
 //   pp. 360-369, 1997
 // http://www.cs.princeton.edu/~rs/strings/paper.pdf
 
+#if defined(TRIE_NEED_GEN_ATTRS) || \
+    defined(TRIE_NEED_GEN_CODE_GET_ATTR) || \
+    defined(TRIE_NEED_GEN_CODE)
+#define TRIE_IMPL_ATTRS_
+#endif
+
 #ifndef TRIE_NEED_IMPL_ONLY
 
+#ifdef TRIE_IMPL_ATTRS_
+struct TRIE_ATTR_TYPE
+{
+    size_t        width;
+    TRIE_SYM_TYPE last;
+};
+#endif // TRIE_IMPL_ATTRS_
+
 struct TRIE_NODE_TYPE
 {
     TRIE_SYM_TYPE sym;
@@ -309,6 +358,10 @@ struct TRIE_NODE_TYPE
         const struct TRIE_NODE_TYPE* eq;
         TRIE_VAL_TYPE                val;
     } cell;
+
+#ifdef TRIE_IMPL_ATTRS_
+    struct TRIE_ATTR_TYPE attr;
+#endif
 };
 
 typedef struct TRIE_NODE_TYPE* (*TRIE_ALLOC_FUNC_TYPE)(
@@ -319,6 +372,9 @@ struct TRIE_TYPE
     TRIE_ALLOC_FUNC_TYPE         alloc_func;
     TRIE_ALLOC_OBJ_TYPE*         alloc_obj;
     const struct TRIE_NODE_TYPE* root;
+#ifdef TRIE_IMPL_ATTRS_
+    struct TRIE_ATTR_TYPE        attr;
+#endif
 };
 
 #endif // TRIE_NEED_IMPL_ONLY
@@ -672,16 +728,15 @@ static bool TRIE_LOOKUP_SYM_NODE(
 #endif // TRIE_NEED_LOOKUP_SYM_NODE ||
        // TRIE_NEED_NODE_HAS_NULL_SIB
 
-#ifdef TRIE_NEED_LOOKUP_SYM
+#if defined(TRIE_NEED_LOOKUP_SYM_AT) || \
+    defined(TRIE_NEED_LOOKUP_SYM)
 
-static bool TRIE_LOOKUP_SYM(
-    const struct TRIE_TYPE* trie, TRIE_SYM_TYPE sym,
+static bool TRIE_LOOKUP_SYM_AT(
+    const struct TRIE_NODE_TYPE* node, TRIE_SYM_TYPE sym,
     const struct TRIE_NODE_TYPE** result)
 {
-    const struct TRIE_NODE_TYPE* node;
     enum trie_sym_cmp_t cmp;
 
-    node = trie->root;
     while (node) {
         cmp = TRIE_SYM_CMP(sym, node->sym); 
         if (cmp == trie_sym_cmp_eq) {
@@ -706,6 +761,18 @@ static bool TRIE_LOOKUP_SYM(
     return false;
 }
 
+#endif // TRIE_NEED_LOOKUP_SYM_AT ||
+       // TRIE_NEED_LOOKUP_SYM
+
+#ifdef TRIE_NEED_LOOKUP_SYM
+
+static inline bool TRIE_LOOKUP_SYM(
+    const struct TRIE_TYPE* trie, TRIE_SYM_TYPE sym,
+    const struct TRIE_NODE_TYPE** result)
+{
+    return TRIE_LOOKUP_SYM_AT(trie->root, sym, result);
+}
+
 #endif // TRIE_NEED_LOOKUP_SYM
 
 #if defined(TRIE_NEED_INSERT_SYM) || \
@@ -763,16 +830,15 @@ done:
 
 #endif // TRIE_NEED_INSERT_SYM
 
-#ifdef TRIE_NEED_LOOKUP_KEY
+#if defined(TRIE_NEED_LOOKUP_KEY_AT) || \
+    defined(TRIE_NEED_LOOKUP_KEY)
 
-static bool TRIE_LOOKUP_KEY(
-    const struct TRIE_TYPE* trie, TRIE_KEY_TYPE key,
+static bool TRIE_LOOKUP_KEY_AT(
+    const struct TRIE_NODE_TYPE* node, TRIE_KEY_TYPE key,
     const struct TRIE_NODE_TYPE** result)
 {
-    const struct TRIE_NODE_TYPE* node;
     enum trie_sym_cmp_t cmp;
 
-    node = trie->root;
     while (node) {
         cmp = TRIE_SYM_CMP(
                 TRIE_KEY_DEREF(key),
@@ -799,8 +865,22 @@ static bool TRIE_LOOKUP_KEY(
     return false;
 }
 
+#endif // TRIE_NEED_LOOKUP_KEY_AT ||
+       // TRIE_NEED_LOOKUP_KEY
+
+#ifdef TRIE_NEED_LOOKUP_KEY
+
+static inline bool TRIE_LOOKUP_KEY(
+    const struct TRIE_TYPE* trie, TRIE_KEY_TYPE key,
+    const struct TRIE_NODE_TYPE** result)
+{
+    return TRIE_LOOKUP_KEY_AT(trie->root, key, result);
+}
+
 #endif // TRIE_NEED_LOOKUP_KEY
 
+#ifdef TRIE_NEED_INSERT_KEY
+
 static bool TRIE_INSERT_KEY(
     struct TRIE_TYPE* trie, TRIE_KEY_TYPE key,
     const struct TRIE_NODE_TYPE** result)
@@ -845,6 +925,8 @@ static bool TRIE_INSERT_KEY(
     return true;
 }
 
+#endif // TRIE_NEED_INSERT_KEY
+
 #define TRIE_NODE_CONST_CAST(p) \
     CONST_CAST(p, struct TRIE_NODE_TYPE)
 
@@ -1204,7 +1286,8 @@ static size_t TRIE_NODE_GET_ITER_MAX_DEPTH(
 #endif // TRIE_NEED_NODE_GET_ITER_MAX_DEPTH ||
        // TRIE_NEED_ITERATOR
 
-#ifdef TRIE_NEED_PRINT
+#if defined(TRIE_NEED_PRINT_NODE) || \
+    defined(TRIE_NEED_PRINT)
 
 static void TRIE_PRINT_NODE(
     const struct TRIE_NODE_TYPE* node, FILE* file)
@@ -1243,6 +1326,11 @@ static void TRIE_PRINT_NODE(
     fputc('}', file);
 }
 
+#endif // TRIE_NEED_PRINT_NODE ||
+       // TRIE_NEED_PRINT
+
+#ifdef TRIE_NEED_PRINT
+
 static void TRIE_PRINT(
     const struct TRIE_TYPE* trie, FILE* file)
 {
@@ -2130,6 +2218,501 @@ static void TRIE_LVL_ITERATOR_DONE(
 
 #endif // TRIE_NEED_LVL_ITERATOR
 
+#ifdef TRIE_IMPL_ATTRS_
+
+#define TRIE_SYM_LT(x, y)     \
+    (                         \
+        TRIE_SYM_CMP(x, y) == \
+        trie_sym_cmp_lt       \
+    )
+
+#define TRIE_ATTR(l, w)       \
+    (struct TRIE_ATTR_TYPE) { \
+        .last = l,            \
+        .width = w            \
+    }
+
+#define TRIE_ATTR_ADD(x, y)              \
+    do {                                 \
+        STATIC(TYPEOF_IS(x,              \
+            struct TRIE_ATTR_TYPE));     \
+        STATIC(TYPEOF_IS(y,              \
+            struct TRIE_ATTR_TYPE));     \
+        SIZE_ADD_EQ(x.width, y.width);   \
+        if (TRIE_SYM_LT(x.last, y.last)) \
+            x.last = y.last;             \
+    } while (0)
+
+static struct TRIE_ATTR_TYPE
+    TRIE_NODE_GEN_ATTRS(
+        const struct TRIE_NODE_TYPE* node)
+{
+    struct TRIE_ATTR_TYPE r =
+           TRIE_ATTR(TRIE_NULL_SYM(), 0);
+
+    ASSERT(node != NULL);
+
+    if (node->lo != NULL)
+        TRIE_ATTR_ADD(r,
+            TRIE_NODE_GEN_ATTRS(node->lo));
+
+    if (!TRIE_SYM_IS_NULL(node->sym) &&
+        node->cell.eq != NULL) {
+        struct TRIE_ATTR_TYPE a =
+               TRIE_ATTR(node->sym, 1);
+
+        TRIE_NODE_CONST_CAST(node)->attr =
+            TRIE_NODE_GEN_ATTRS(node->cell.eq);
+
+        TRIE_ATTR_ADD(r, a);
+    }
+
+    if (node->hi != NULL)
+        TRIE_ATTR_ADD(r,
+            TRIE_NODE_GEN_ATTRS(node->hi));
+
+    return r;
+}
+
+#endif // TRIE_IMPL_ATTRS_
+
+#define TRIE_NODE_N_CHILD(n)          \
+    (                                 \
+        !TRIE_SYM_IS_NULL((n)->sym) + \
+        ((n)->lo != NULL) +           \
+        ((n)->hi != NULL)             \
+    )
+#define TRIE_NODE_VALUE_CHILD(n)               \
+    ({                                         \
+        const struct TRIE_NODE_TYPE* __r = n;  \
+        ASSERT(__r != NULL);                   \
+        do {                                   \
+            /* invariant:                   */ \
+            /*  TRIE_SYM_IS_NULL(__r->sym)  */ \
+            /*  => __r->lo == NULL          */ \
+            ASSERT(                            \
+                !TRIE_SYM_IS_NULL(__r->sym) || \
+                __r->lo == NULL);              \
+        }                                      \
+        while (!TRIE_SYM_IS_NULL(__r->sym) &&  \
+               (__r = __r->lo) != NULL);       \
+        __r;                                   \
+    })
+#define TRIE_NODE_VALUE(n)                  \
+    ({                                      \
+        ASSERT(TRIE_SYM_IS_NULL((n)->sym)); \
+        (n)->cell.val;                      \
+    })
+
+#if defined(TRIE_NEED_GEN_CODE_GET_ATTR) || \
+    defined(TRIE_NEED_GEN_CODE)
+
+#ifndef TRIE_GEN_CODE_ATTRS
+#define TRIE_GEN_CODE_ATTRS
+enum trie_gen_code_attrs_t
+{
+    trie_gen_code_attrs_has_prefix = 1 << 0,
+    trie_gen_code_attrs_has_equal  = 1 << 1
+};
+#endif
+
+#ifndef TRIE_GEN_CODE_OPTS
+#define TRIE_GEN_CODE_OPTS
+struct trie_gen_code_opts_t
+{
+    bits_t      expanded: 1;
+    const char* prefix;
+    const char* equal;
+};
+#endif
+
+#endif // TRIE_NEED_GEN_CODE_GET_ATTR) ||
+       // TRIE_NEED_GEN_CODE
+
+#ifdef TRIE_NEED_GEN_CODE_GET_ATTR
+
+// stev: the 'TRIE_{NODE_,}GEN_CODE_GET_ATTR'
+// functions below are mere reductions of the
+// functions 'TRIE_{NODE_,}GEN_CODE'
+
+static enum trie_gen_code_attrs_t
+    TRIE_NODE_GEN_CODE_GET_ATTR(
+        const struct TRIE_NODE_TYPE* node,
+        const struct TRIE_ATTR_TYPE* attr,
+        size_t* path, size_t start)
+{
+    enum trie_gen_code_attrs_t r = 0;
+
+    ASSERT(node != NULL);
+
+    if (node->lo != NULL)
+        r |= TRIE_NODE_GEN_CODE_GET_ATTR(
+                node->lo, attr, path, start);
+
+    if (!TRIE_SYM_IS_NULL(node->sym) &&
+        node->cell.eq != NULL) {
+        const size_t n =
+            TRIE_NODE_N_CHILD(node->cell.eq);
+        const struct TRIE_NODE_TYPE* child =
+            TRIE_NODE_VALUE_CHILD(node->cell.eq);
+        const bool at =
+            !child && n == 1;
+        const bool obj =
+            n && !at;
+
+        const size_t ps = *path;
+        const size_t l = SIZE_SUB(ps, start);
+
+        const bool b = l > 0;
+        const bool c =
+            !child || node->attr.width;
+        const bool w = attr->width > 1;
+        const size_t s = start + (w && b);
+        const bool t = !w || b;
+        const size_t m = ps - s + t;
+        const bool i = t || s < ps;
+        const bool j = c && i;
+        const bool v = m <= 1;
+
+        // stev: s <= ps:
+        // case 1: l == 0 => s = start = ps
+        // case 2: l > 0  => start < ps
+        //                => s <= start + 1 <= ps
+        ASSERT(s <= ps);
+
+        if (at) (*path) ++;
+
+        if (!at && !v)
+            r |= j ? trie_gen_code_attrs_has_prefix
+                   : trie_gen_code_attrs_has_equal;
+
+        r |= TRIE_NODE_GEN_CODE_GET_ATTR(
+                node->cell.eq,
+                obj ? &node->attr : attr,
+                path, !at ? ps : start);
+
+        if (at) (*path) --;
+    }
+
+    if (node->hi != NULL)
+        r |= TRIE_NODE_GEN_CODE_GET_ATTR(
+                node->hi, attr, path, start);
+
+    return r;
+}
+
+static enum trie_gen_code_attrs_t
+    TRIE_GEN_CODE_GET_ATTR(
+        struct TRIE_TYPE* trie,
+        const struct trie_gen_code_opts_t* opts)
+{
+    size_t path = 0;
+
+    // stev: root == NULL does have
+    // meaning: the trie is empty
+    if (trie->root == NULL ||
+        opts->expanded)
+        return 0;
+
+    trie->attr = TRIE_NODE_GEN_ATTRS(
+        trie->root);
+
+    return TRIE_NODE_GEN_CODE_GET_ATTR(
+        trie->root, &trie->attr, &path, 0);
+}
+
+#endif // TRIE_NEED_GEN_CODE_GET_ATTR
+
+#ifdef TRIE_NEED_GEN_CODE
+
+#undef  STACK_NAME
+#define STACK_NAME TRIE_PATH_STACK_NAME
+
+#undef  STACK_ELEM_TYPE
+#define STACK_ELEM_TYPE TRIE_SYM_TYPE
+
+#include "stack-impl.h"
+
+#define TRIE_PATH_TYPE TRIE_PATH_STACK_TYPE
+#define TRIE_PATH_INIT TRIE_PATH_STACK_INIT
+#define TRIE_PATH_DONE TRIE_PATH_STACK_DONE
+#define TRIE_PATH_NAME TRIE_PATH_STACK_NAME
+
+#define TRIE_PATH_PUSH(s) STACK_PUSH(path, s)
+#define TRIE_PATH_POP()   STACK_POP(path)
+
+#define TRIE_GEN_CODE_PRINT_SYM(s, q) \
+    TRIE_SYM_GEN_CODE(s, q, file)
+#define TRIE_GEN_CODE_PRINT_VAL(v) \
+    TRIE_VAL_GEN_CODE(v, file)
+#define TRIE_GEN_CODE_PRINT_ERR() \
+    TRIE_ERR_GEN_CODE(file)
+
+#define TRIE_GEN_CODE_PRINT_FMT(l, s, ...) \
+    do {                                   \
+        fprintf(file, "%*s" s,             \
+            SIZE_AS_INT(SZ(4) * (l)), "",  \
+            ## __VA_ARGS__);               \
+    } while (0)
+#define TRIE_GEN_CODE_PRINT_INDENT(l) \
+    TRIE_GEN_CODE_PRINT_FMT(l, )
+#define TRIE_GEN_CODE_PRINT_IF(c, s) \
+    do {                             \
+        if (c) fputs(s, file);       \
+    } while (0)
+#define TRIE_GEN_CODE_PRINT_CH(c) \
+    do {                          \
+        fputc(c, file);           \
+    } while (0)
+#define TRIE_GEN_CODE_PRINT(s) \
+    do {                       \
+        fputs(s, file);        \
+    } while (0)
+
+// stev: the 'TRIE_{NODE_,}GEN_CODE' functions
+// below are due to my work on Trie-Gen project,
+// http://www.nongnu.org/trie-gen/; they are C
+// transcriptions of the C++ code in source file
+// 'src/trie.hpp' of Trie-Gen -- more precisely,
+// of function 'TernaryTrie<>::gen_compact_code'
+// and a few prerequisites of it; the algorithm
+// generating trie lookup functions constitutes
+// original work (of course, modulo my knowledge
+// of the literature of trie data structures)
+
+static enum trie_gen_code_attrs_t
+    TRIE_NODE_GEN_CODE(
+        const struct TRIE_NODE_TYPE* node,
+        const struct trie_gen_code_opts_t* opts,
+        const struct TRIE_ATTR_TYPE* attr,
+        struct TRIE_PATH_TYPE* path,
+        size_t start, size_t level,
+        FILE* file)
+{
+    enum trie_gen_code_attrs_t r = 0;
+
+    ASSERT(node != NULL);
+
+    if (node->lo != NULL)
+        r |= TRIE_NODE_GEN_CODE(
+                node->lo, opts, attr, path,
+                start, level, file);
+
+    if (!TRIE_SYM_IS_NULL(node->sym) &&
+        node->cell.eq != NULL) {
+        const size_t n =
+            TRIE_NODE_N_CHILD(node->cell.eq);
+        const struct TRIE_NODE_TYPE* child =
+            TRIE_NODE_VALUE_CHILD(node->cell.eq);
+        const bool at =
+            !child && n == 1;
+        const bool obj =
+            n && !at;
+
+        const TRIE_SYM_TYPE* pd = STACK_DATA(path);
+        const size_t ps = STACK_SIZE(path);
+        const size_t l = SIZE_SUB(ps, start);
+
+        const bool b = l > 0;
+        const bool c =
+            !child || node->attr.width;
+        const bool w = attr->width > 1;
+        const size_t s = start + (w && b);
+        const bool t = !w || b;
+        const size_t m = ps - s + t;
+        const bool i = t || s < ps;
+        const bool j = c && i;
+        const size_t d = 1 + (w && b);
+        const bool a = c || !i;
+        const bool v = m <= 1 ||
+            opts->expanded;
+
+        // stev: s <= ps:
+        // case 1: l == 0 => s = start = ps
+        // case 2: l > 0  => start < ps
+        //                => s <= start + 1 <= ps
+        ASSERT(s <= ps);
+
+        if (at)
+            TRIE_PATH_PUSH(node->sym);
+
+        // stev: !at <=> child || n != 1;
+        //       therefore: child => !at
+        if (!at) {
+            const TRIE_SYM_TYPE *p, *e;
+
+            if (w) {
+                TRIE_SYM_TYPE s =
+                    b ? pd[start] : node->sym;
+
+                TRIE_GEN_CODE_PRINT_FMT(
+                    level, "case ");
+                TRIE_GEN_CODE_PRINT_SYM(
+                    s, true);
+                TRIE_GEN_CODE_PRINT(":\n");
+            }
+            if (i)
+                TRIE_GEN_CODE_PRINT_FMT(
+                    level + w, "if (");
+            if (!v) {
+                r |= j ? trie_gen_code_attrs_has_prefix
+                       : trie_gen_code_attrs_has_equal;
+                TRIE_GEN_CODE_PRINT(
+                    j ? opts->prefix : opts->equal);
+                TRIE_GEN_CODE_PRINT_CH('(');
+                TRIE_GEN_CODE_PRINT_IF(!j, "p, ");
+                TRIE_GEN_CODE_PRINT_CH('"');
+            }
+            for (p = pd + s,
+                 e = pd + ps;
+                 p < e;
+                 p ++) {
+                if (v)
+                    TRIE_GEN_CODE_PRINT(
+                        "*p ++ == ");
+                if (true)
+                    TRIE_GEN_CODE_PRINT_SYM(
+                        *p, v);
+                if (v && (p + 1 < e || t)) {
+                    TRIE_GEN_CODE_PRINT(
+                        " &&\n");
+                    TRIE_GEN_CODE_PRINT_INDENT(
+                        level + w + 1);
+                }
+            }
+            if (t) {
+                if (v)
+                    TRIE_GEN_CODE_PRINT(
+                        "*p ++ == ");
+                if (true)
+                    TRIE_GEN_CODE_PRINT_SYM(
+                        node->sym, v);
+            }
+            if (!v) {
+                TRIE_GEN_CODE_PRINT_CH('"');
+                TRIE_GEN_CODE_PRINT_IF(j, ", p");
+                TRIE_GEN_CODE_PRINT_CH(')');
+            }
+            if (j)
+                TRIE_GEN_CODE_PRINT(") {\n");
+            if (!v && j)
+                TRIE_GEN_CODE_PRINT_FMT(
+                    level + w + 1,  "p += %zu;\n",
+                    m);
+        }
+
+        // stev: child => !at (see above)
+        if (child) {
+            TRIE_VAL_TYPE u =
+                TRIE_NODE_VALUE(child);
+
+            if (a)
+                TRIE_GEN_CODE_PRINT_FMT(
+                    level + d, "if (");
+            else
+            if (v || j) {
+                TRIE_GEN_CODE_PRINT(
+                    " &&\n");
+                TRIE_GEN_CODE_PRINT_INDENT(
+                    level + d);
+            }
+
+            TRIE_GEN_CODE_PRINT_IF(
+                v || j, "*p == 0");
+            TRIE_GEN_CODE_PRINT(
+                ")\n");
+            TRIE_GEN_CODE_PRINT_FMT(
+                level + d + a, "return ");
+            TRIE_GEN_CODE_PRINT_VAL(u);
+            TRIE_GEN_CODE_PRINT(";\n");
+        }
+
+        if (obj && node->attr.width > 1)
+            TRIE_GEN_CODE_PRINT_FMT(
+                level + d, "switch (*p ++) {\n");
+
+        r |= TRIE_NODE_GEN_CODE(
+                node->cell.eq, opts,
+                obj ? &node->attr : attr,
+                path, !at ? ps : start,
+                level + (obj ? d : 0),
+                file);
+
+        if (obj && node->attr.width > 1)
+            TRIE_GEN_CODE_PRINT_FMT(
+                level + d, "}\n");
+
+        if (!at && j)
+            TRIE_GEN_CODE_PRINT_FMT(
+                level + w, "}\n");
+
+        if (attr->last != node->sym && !l) {
+            TRIE_GEN_CODE_PRINT_FMT(
+                level + 1, "return ");
+            TRIE_GEN_CODE_PRINT_ERR();
+            TRIE_GEN_CODE_PRINT(";\n");
+        }
+
+        if (at)
+            TRIE_PATH_POP();
+    }
+
+    if (node->hi != NULL)
+        r |= TRIE_NODE_GEN_CODE(
+                node->hi, opts, attr, path,
+                start, level, file);
+
+    return r;
+}
+
+static enum trie_gen_code_attrs_t
+    TRIE_GEN_CODE(
+        struct TRIE_TYPE* trie,
+        const struct trie_gen_code_opts_t* opts,
+        FILE* file)
+{
+    enum trie_gen_code_attrs_t r = 0;
+
+    if (trie->root != NULL)
+        trie->attr = TRIE_NODE_GEN_ATTRS(
+            trie->root);
+
+    if (trie->attr.width > 1)
+        TRIE_GEN_CODE_PRINT_FMT(
+            1, "switch (*p ++) {\n");
+
+    // stev: root == NULL does have
+    // meaning: the trie is empty
+    if (trie->root != NULL) {
+        struct TRIE_PATH_TYPE path;
+
+        TRIE_PATH_INIT(&path, 0);
+
+        r |= TRIE_NODE_GEN_CODE(
+                trie->root, opts,
+                &trie->attr,
+                &path, 0, 1,
+                file);
+
+        TRIE_PATH_DONE(&path);
+    }
+
+    if (trie->attr.width > 1)
+        TRIE_GEN_CODE_PRINT_FMT(
+            1, "}\n");
+
+    if (true) {
+        TRIE_GEN_CODE_PRINT_FMT(
+            1, "return ");
+        TRIE_GEN_CODE_PRINT_ERR();
+        TRIE_GEN_CODE_PRINT(";\n");
+    }
+
+    return r;
+}
+
+#endif // TRIE_NEED_GEN_CODE
+
 #endif // TRIE_NEED_STRUCT_ONLY
 
 #ifndef __TRIE_IMPL_H
