@@ -1,4 +1,4 @@
-// Copyright (C) 2016, 2017, 2018  Stefan Vargyas
+// Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas
 // 
 // This file is part of Json-Type.
 // 
@@ -23,6 +23,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <alloca.h>
 #include <signal.h>
 #include <unistd.h>
 #include <errno.h>
@@ -34,7 +35,9 @@
 #include <link.h>
 
 #include "lib/json.h"
+#include "lib/json-filter.h"
 #include "lib/pretty-print.h"
+#include "lib/su-size.h"
 
 #include "common.h"
 #include "char-traits.h"
@@ -52,10 +55,10 @@
 
 static const char program[] = STRINGIFY(PROGRAM);
 static const char library[] = STRINGIFY(LIBRARY);
-static const char verdate[] = "0.8.10 -- 2018-04-26 11:22"; // $ date +'%F %R'
+static const char verdate[] = "0.9.0 -- 2018-06-30 12:53"; // $ date +'%F %R'
 
 static const char license[] =
-"Copyright (C) 2016, 2017, 2018  Stefan Vargyas.\n"
+"Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas.\n"
 "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n"
 "This is free software: you are free to change and redistribute it.\n"
 "There is NO WARRANTY, to the extent permitted by law.\n";
@@ -430,6 +433,14 @@ static void lib_version_check(void)
 #error "JSON_VERSION has an invalid value"
 #endif
 
+#ifndef JSON_DEBUG
+#define JSON_DEBUG_STATE "disabled"
+#else
+#define JSON_DEBUG_STATE "enabled"
+#endif
+
+#define FILTER_VERSION VERSION_ID(JSON_FILTER)
+
 static void version(void)
 {
     struct lib_version_t v;
@@ -441,27 +452,21 @@ static void version(void)
         "%s: %s: version: %u.%u.%u\n"
         "%s: %s: build: %s\n"
         "%s: %s: path: %s\n"
-        "%s: %s: JSON_DEBUG is %sabled\n\n%s",
+        "%s: %s: JSON_DEBUG is " JSON_DEBUG_STATE "\n"
+        "%s: json filter version: %s\n\n%s",
         program, verdate,
         program, library, v.major, v.minor, v.patch,
         program, library, v.build,
         program, library, v.rpath
             ? v.rpath : "???",
-        program, library,
-#ifdef JSON_DEBUG
-        "en",
-#else
-        "dis",
-#endif
+        program, library, program,
+        STRINGIFY(FILTER_VERSION),
         license
     );
 
     lib_version_done(&v);
 }
 
-#define KB(X) (SZ(X) * SZ(1024))
-#define MB(X) (KB(X) * SZ(1024))
-
 #define PARSE_BUF_SIZE_OPTARG   parse_su_size_optarg
 #define PARSE_STACK_SIZE_OPTARG parse_su_size_optarg
 #define PARSE_POOL_SIZE_OPTARG  parse_su_size_optarg
@@ -550,7 +555,7 @@ static void version(void)
 static void usage(void)
 {
     fprintf(stdout,
-        "usage: %s [ACTION|OPTION]... [FILE]\n"
+        "usage: %s [ACTION|OPTION]... [FILE] [-- FILTER_LIB [FILTER_ARG]...]...\n"
         "where actions are specified as:\n"
         "  -O|--parse-only              run 'json-parser': only parse the input or\n"
         "  -E|--echo-parse                parse and echo out verbatim the input\n"
@@ -558,7 +563,7 @@ static void usage(void)
         "  -R|--terse-print               tersely the input JSON objects; otherwise\n"
         "  -Y|--type-print                build the AST of the JSON input and print\n"
         "  -A|--ast-print                 out either its type object or the AST or\n"
-        "  -F|--from-ast-print            reprint prettily from AST the JSON input\n"
+        "  -M|--from-ast-print            reprint prettily from AST the JSON input\n"
         "  -T|--typelib-func [NAME]     run 'json-typelib' on the input typelib for\n"
         "                                 the named function:\n"
         "                                   v|validate    check the validity of input;\n"
@@ -574,6 +579,13 @@ static void usage(void)
         "                                   d|gen-def     generate the type def code for\n"
         "                                                 the constructed type\n"
         "                                 the default is validate\n"
+        "  -F|--filterlib-exec          run 'json-filterlib': execute the filter library\n"
+        "                                 specified by the first command line argument\n"
+        "                                 group of form `-- FILTER_LIB [FILTER_ARG]...';\n"
+        "                                 if the invoking command line specifies an input\n"
+        "                                 file name (prior to the first `--'), then take\n"
+        "                                 that to be library name and load the library by\n"
+        "                                 passing to it the argument `--help' only\n"
         "  -S|--escape-json             run 'json-escape': produce the JSON-escaped text\n"
         "                                 of the input UTF8 text given\n"
         "  -J|--json2                   run 'json2': flatten the hierarchical structure\n"
@@ -583,13 +595,21 @@ static void usage(void)
         "the options are:\n"
         "  -d|--type-def TYPE           use the specified type definition or library for\n"
         "  -t|--type-lib FILE[:NAME]      type checking the JSON input or, otherwise, do\n"
-        "  -N|--type-name [NAME]          not check the input for types at all (default);\n"
+        "  -n|--type-name [NAME]          not check the input for types at all (default);\n"
         "     --no-type-check             the optional name specifies which type to use\n"
         "                                 in case the def/library is made of an array\n"
         "                                 of named types instead of just a sole type;\n"
         "                                 type checking doesn't apply yet to the AST\n"
         "                                 options `-Y|--type-print', `-A|--ast-print'\n"
-        "                                 and `-F|--from-ast-print'\n"
+        "                                 and `-M|--from-ast-print'\n"
+        "  -f|--[no-]filter-libs        build a JSON input filtering chain using the\n"
+        "                                 specified filter libraries and arguments or,\n"
+        "                                 otherwise, do no filter the input at all;\n"
+        "                                 the filtering chain is specified as a series\n"
+        "                                 of one or more command line arguments groups\n"
+        "                                 of form `-- FILTER_LIB [FILTER_ARG]...'; the\n"
+        "                                 filtering series is placed immediately after\n"
+        "                                 the main program's command line arguments\n"
         "  -i|--input-file FILE         take input from the named file (default: '-')\n"
 #undef  CASE
 #define CASE(type, id, name, pad, letter, desc, help, min, max, def) \
@@ -605,7 +625,7 @@ static void usage(void)
         "                                 with `-c|--escape-utf8', generates surrogate pairs\n"
         "                                 in the strings produced by `-P|--pretty-print',\n"
         "                                 `-R|--terse-print', `Y|--type-print',\n"
-        "                                 `-A|--ast-print' or `F|--from-ast-print'\n"
+        "                                 `-A|--ast-print' or `M|--from-ast-print'\n"
         "  -u|--[no-]valid[ate]-utf8    validate for UTF8 encoding the non-ascii chars\n"
         "                                 in JSON strings (when the action given is any but\n"
         "                                 `-S|--escape-json') or in whole text (when given\n"
@@ -613,7 +633,7 @@ static void usage(void)
         "  -c|--[no-]escape-utf8        escape UTF8 byte sequences or do not (default not)\n"
         "                                 in the strings produced by `-P|--pretty-print',\n"
         "                                 `-R|--terse-print', `Y|--type-print',\n"
-        "                                 `-A|--ast-print' or `F|--from-ast-print'\n"
+        "                                 `-A|--ast-print' or `M|--from-ast-print'\n"
         "  -a|--[no-]ascii-only         enforce JSON strings to contain ASCII chars\n"
         "                                 only or otherwise do not (default not)\n"
         "  -m|--[no-]multi-object       allow multiple JSON objects on input or\n"
@@ -626,16 +646,24 @@ static void usage(void)
         "                                 or 'ignore' for ignoring the signal altogether\n"
         "                                 (the default is 'error'); option `-p' cuts short\n"
         "                                 the long option `--sigpipe-action=no-error'\n"
-        "  -Q|--[no-]quote-text         print out double quote chars surrounding the\n"
+        "  -q|--[no-]quote-text         print out double quote chars surrounding the\n"
         "                                 text produced by action `-S|--escape-json'\n"
         "                                 or otherwise do not (default not)\n"
         "  -r|--[no-]trim-spaces        trim out heading and trailing space chars in the\n"
         "                                 input text when action is `-S|--escape-json'\n"
         "                                 or otherwise do not (default not)\n"
+        "  -w|--[no-]raw-strings        allow C++11-like raw string literals when JSON\n"
+        "                                 text expected or otherwise do not (default not);\n"
+        "                                 also produce raw string literals if action is\n"
+        "                                 `-P|--pretty-print' or `-M|--from-ast-print';\n"
+        "                                 note that in case of the other 'json-printer'\n"
+        "                                 actions no such literals get produced at all\n"
         "  -y|--[no-]empty-input        allow empty input when JSON text expected or\n"
         "                                 otherwise do not (default not)\n"
         "     --[no-]newline-sep        separate the top values on terse JSON print outs\n"
         "                                 by newline or otherwise do not (default not)\n"
+        "     --[no-]warning            print out warning reports on stderr or otherwise\n"
+        "                                 do not (default do)\n"
         "     --[no-]error              print out error reports on stderr or otherwise\n"
         "                                 do not (default do); note that reporting the\n"
         "                                 fatal errors on stderr is never suppressed\n"
@@ -659,199 +687,81 @@ static void usage(void)
         program);
 }
 
-struct su_size_t
-{
-    size_t sz;
-    const char* su;
-};
-
-static struct su_size_t su_size(size_t sz)
-{
-    struct su_size_t r;
-
-    r.sz = sz;
-    r.su = "";
-    if (r.sz  % MB(1) == 0) {
-        r.sz /= MB(1);
-        r.su = "M";
-    }
-    else
-    if (r.sz  % KB(1) == 0) {
-        r.sz /= KB(1);
-        r.su = "K";
-    }
-    return r;
-}
-
-static size_t null_size_to_string(
-    char* buf, size_t len)
-{
-    ASSERT(len >= 2);
-    strncpy(buf, "-", 2);
-    return 1;
-} 
-
-static size_t su_size_to_string(
-    char* buf, size_t len, size_t val, bool null)
-{
-    if (val == 0 && null)
-        return null_size_to_string(buf, len);
-    else {
-        struct su_size_t s;
-        size_t r;
-        int n;
-
-        s = su_size(val);
-        n = snprintf(buf, len, "%zu%s", s.sz, s.su);
-        r = INT_AS_SIZE(n);
-
-        ASSERT(n >= 0);
-        ASSERT(r < len);
-
-        return r;
-    }
-}
-
-static size_t size_to_string(
-    char* buf, size_t len, size_t val, bool null)
-{
-    if (val == 0 && null)
-        return null_size_to_string(buf, len);
-    else {
-        size_t r;
-        int n;
-
-        n = snprintf(buf, len, "%zu", val);
-        r = INT_AS_SIZE(n);
-
-        ASSERT(n >= 0);
-        ASSERT(r < len);
-
-        return r;
-    }
-}
-
 static void usage_sizes(void)
 {
-    enum sizes_param_type_t {
-        NONE, BUF, STACK, POOL, WIDTH
+    enum {
+        BUF   = su_size_param_su,
+        STACK = su_size_param_su,
+        POOL  = su_size_param_su,
+        WIDTH = su_size_param_plain,
     };
-    typedef size_t (*to_string_func_t)(
-        char*, size_t, size_t, bool);
-    static const to_string_func_t to_strings[] = {
-        [BUF]   = su_size_to_string,
-        [STACK] = su_size_to_string,
-        [POOL]  = su_size_to_string,
-        [WIDTH] = size_to_string,
-    };
-    struct sizes_param_t
-    {
-        enum sizes_param_type_t
-                    type;
-        const char* name;
-        const char* desc;
-        size_t      min;
-        size_t      max;
-        size_t      def;
-    };
-    static const struct sizes_param_t params[] = {
-        { NONE, "name", "description", 0, 0, 0 },
+    static const struct su_size_param_t params[] = {
 #undef  CASE
 #define CASE(type, id, name, pad, letter, desc, help, min, max, def) \
         { type, name, desc, min, max, def },
 #include "opt-sizes.def"
     };
+
+    su_size_print_sizes(params, ARRAY_SIZE(params), stdout);
+}
+
+static char* dump_filters(
+    const struct options_filter_t* filters,
+    size_t n_filters, const char* name)
+{
     const size_t m = TYPE_DIGITS10(size_t) + 1;
-    struct sizes_values_t
-    {
-        char min[m];
-        char max[m];
-        char def[m];
-    };
-    const size_t n = ARRAY_SIZE(params);
-    struct sizes_values_t values[n];
-    struct sizes_max_widths_t
-    {
-        size_t name;
-        size_t min;
-        size_t max;
-        size_t def;
-    };
-    struct sizes_max_widths_t w;
-    const struct sizes_param_t *p, *e;
-    struct sizes_values_t *v, *f;
+    const struct options_filter_t* p;
+    size_t i, n = 0, l, s = 2;
+    const size_t w = 31;
+    char *b = NULL, *d;
+    FILE* f;
 
-    memset(values, 0, sizeof(values));
-    memset(&w, 0, sizeof(w));
+    l = strlen(name);
 
-    for (p = params,
-         v = values,
-         e = params + n,
-         f = values + n;
-         p < e && v < f;
-         p ++,
-         v ++) {
-        size_t l;
+    f = open_memstream(&b, &n);
+    VERIFY(f != NULL);
 
-        if (v == values) {
-            ASSERT(m >= 8);
+    if (n_filters == 0) {
+        fprintf(f, "%s:%*s-\n", name, w - 1  > l
+            ? SIZE_AS_INT((w - 1) - l) : 0, "");
 
-            strncpy(v->min, "min", 4);
-            strncpy(v->max, "max", 4);
-            strncpy(v->def, "default", 8);
-        }
-        else {
-            to_string_func_t f =
-                ARRAY_NULL_ELEM(to_strings, p->type);
-            ASSERT(f != NULL);
-
-            f(v->min, m, p->min, true);
-            f(v->max, m, p->max, true);
-            f(v->def, m, p->def, false);
-        }
-
-        l = strlen(p->name);
-        if (w.name < l)
-            w.name = l;
-
-        l = strlen(v->min);
-        if (w.min < l)
-            w.min = l;
-
-        l = strlen(v->max);
-        if (w.max < l)
-            w.max = l;
-
-        l = strlen(v->def);
-        if (w.def < l)
-            w.def = l;
+        goto done;
     }
 
-    for (p = params,
-         v = values,
-         e = params + n,
-         f = values + n;
-         p < e && v < f;
-         p ++,
-         v ++) {
-        fprintf(stdout,
-            "%-*s  %*s  %*s  %*s  %s\n",
-            SIZE_AS_INT(w.name), p->name,
-            SIZE_AS_INT(w.min), v->min,
-            SIZE_AS_INT(w.max), v->max,
-            SIZE_AS_INT(w.def), v->def,
-            p->desc);
+    SIZE_ADD_EQ(s, l);
+    SIZE_ADD_EQ(s, m);
+    d = alloca(s);
+
+    for (i = 0,
+         p = filters;
+         i < n_filters;
+         i ++,
+         p ++) {
+        int c;
+
+        c = snprintf(d, s, "%s[%zu]", name, i);
+        n = VERIFY_INT_AS_SIZE(c);
+        ASSERT(n < s);
+
+        pretty_print_strings(
+            f, PTR_PTR_CAST(p->argv, char),
+            p->argc, d, w, 0);
     }
+
+done:
+    fclose(f);
+    return b;
 }
 
 static void dump_options(const struct options_t* opts)
 {
     static const char* const objects[] = {
-        [OBJ_JSON_PARSER]  = "json-parser",
-        [OBJ_JSON_PRINTER] = "json-printer",
-        [OBJ_JSON_TYPELIB] = "json-typelib",
-        [OBJ_JSON_ESCAPE]  = "json-escape",
-        [OBJ_JSON2]        = "json2"
+        [OBJ_JSON_PARSER]    = "json-parser",
+        [OBJ_JSON_PRINTER]   = "json-printer",
+        [OBJ_JSON_TYPELIB]   = "json-typelib",
+        [OBJ_JSON_FILTERLIB] = "json-filterlib",
+        [OBJ_JSON_ESCAPE]    = "json-escape",
+        [OBJ_JSON2]          = "json2"
 //!!!2JSON		[OBJ_2JSON]        = "2json"
     };
 #undef  CASE
@@ -906,8 +816,15 @@ static void dump_options(const struct options_t* opts)
 #define CASE(type, id, name, pad, letter, desc, help, min, max, def) \
     SU_SIZE(id, type)
 #include "opt-sizes.def"
-    char* const* p;
-    size_t i;
+    char* b;
+    bool f;
+
+    f = opts->filter_libs ||
+        opts->object == OBJ_JSON_FILTERLIB;
+    b = dump_filters(
+            f ? opts->filters : NULL,
+            f ? opts->n_filters : 0,
+            "filter-libs");
 
     fprintf(stdout,
         "input:                         %s\n"
@@ -918,9 +835,10 @@ static void dump_options(const struct options_t* opts)
         "type-check:                    %s\n"
         "type-def:                      %s\n"
         "type-lib:                      %s\n"
-        "type-name:                     %s\n"
+        "type-name:                     %s\n%s"
         "sigpipe-act:                   %s\n"
         "empty-input:                   %s\n"
+        "raw-strings:                   %s\n"
         "liter-value:                   %s\n"
         "surrogate-pairs:               %s\n"
         "valid-utf8:                    %s\n"
@@ -931,6 +849,7 @@ static void dump_options(const struct options_t* opts)
         "quote-text:                    %s\n"
         "trim-spaces:                   %s\n"
         "newline-sep:                   %s\n"
+        "no-warning:                    %s\n"
         "no-error:                      %s\n"
         "verbose:                       %s\n"
 #undef  CASE
@@ -965,8 +884,10 @@ static void dump_options(const struct options_t* opts)
         NNUL(type_def),
         NNUL(type_lib),
         NNUL(type_name),
+        b != NULL ? b : "",
         NAME(sigpipe_act),
         NOYES(empty_input),
+        NOYES(raw_strings),
         NOYES(liter_value),
         NOYES(surrogate_pairs),
         NOYES(valid_utf8),
@@ -977,6 +898,7 @@ static void dump_options(const struct options_t* opts)
         NOYES(quote_text),
         NOYES(trim_spaces),
         NOYES(newline_sep),
+        NOYES(no_warning),
         NOYES(no_error),
         NOYES(verbose),
 #undef  CASE
@@ -999,23 +921,11 @@ static void dump_options(const struct options_t* opts)
 #undef  NAME0
         opts->argc);
 
-    for (i = 0,
-         p = opts->argv;
-         i < opts->argc;
-         i ++,
-         p ++) {
-        const size_t m = TYPE_DIGITS10(size_t) + 1;
-        const size_t w = 24;
-        char b[m];
-        size_t n;
+    free(b);
 
-        n = size_to_string(b, m, i, false);
-        ASSERT(n < m);
-
-        fprintf(stdout,
-            "argv[%s]:%*s%s\n",
-            b, w > n ? SIZE_AS_INT(w - n) : 0, "", *p);
-    }
+    pretty_print_strings(stdout,
+        PTR_CONST_PTR_CAST(opts->argv, char),
+        opts->argc, "argv", 31, 0);
 }
 
 static void invalid_opt_arg(const char* opt_name, const char* opt_arg)
@@ -1062,6 +972,8 @@ static const char* parse_typelib_optarg(
         illegal_opt_arg(opt_name, opt_arg);
 
     if ((p = strrchr(opt_arg, ':'))) {
+        if (p[1] == '\0')
+            illegal_opt_arg(opt_name, opt_arg);
         *p ++ = 0;
         *type_name = p;
     }
@@ -1069,6 +981,77 @@ static const char* parse_typelib_optarg(
     return opt_arg;
 }
 
+static bool find_filter_sep(
+    const struct options_filter_t* base,
+    size_t* pos)
+{
+    char *const *p, *const* e;
+
+    for (p = base->argv,
+         e = p + base->argc;
+         p < e;
+         p ++) {
+        if (!strcmp(*p, "--")) {
+            *pos = PTR_DIFF(p, base->argv);
+            return true;
+        }
+    }
+    return false;
+}
+
+static bool update_filter_opt(
+    struct options_filter_t* base,
+    struct options_filter_t* prev)
+{
+    size_t k;
+
+    if (!find_filter_sep(base, &k))
+        return false;
+
+    ASSERT(k < base->argc);
+
+    prev->argv = base->argv;
+    prev->argv[k] = NULL;
+    prev->argc = k ++;
+
+    base->argc -= k;
+    base->argv += k;
+
+    return true;
+}
+
+
+static void parse_filters(
+    struct options_filter_t base,
+    struct options_filter_t* libs,
+    size_t* n_libs, size_t max)
+{
+    size_t n = 1;
+
+    ASSERT(max > 0);
+
+    while (true) {
+        if (base.argc == 0)
+        empty_args:
+            error("empty filter arguments "
+                "group #%zu", n);
+        if (n > max)
+            error("too many filter arguments "
+                "groups (max is %zu)", max);
+
+        if (!update_filter_opt(&base, libs)) {
+            *libs = base;
+            break;
+        }
+        if (libs->argc == 0)
+            goto empty_args;
+
+        libs ++;
+        n ++;
+    }
+    *n_libs = n;
+}
+
 // $ . ~/lookup-gen/commands.sh
 // $ print() { printf '%s\n' "$@"; }
 // $ print 'v validate =options_validate_typelib_func' 'p print =options_print_typelib_func' 'a attr =options_attr_typelib_func' 'c check check-attr =options_check_typelib_func' 'A print-check print-check-attr =options_print_check_typelib_func' 'd gen-def =options_def_typelib_func'|gen-func -f lookup_typelib_func_name|ssed -R '1s/^/static /;1s/\s*result_t\&/\n\tenum options_typelib_func_t*/;s/(?=t =)/*/;s/result_t:://'
@@ -1266,78 +1249,10 @@ static enum options_sigpipe_act_t parse_sigpipe_act_optarg(
     return r;
 }
 
-static size_t parse_num(
-    const char* p, const char** q, int b)
-{
-    if (!ISDIGIT(*p)) {
-        *q = p;
-        errno = EINVAL;
-        return 0;
-    }
-    errno = 0;
-    STATIC(SIZE_MAX == ULONG_MAX);
-    return strtoul(p, (char**) q, b);
-}
-
-static size_t parse_size_optarg(
-    const char* opt_name, const char* opt_arg,
-    size_t min, size_t max)
-{
-    const char *p, *q = NULL;
-    size_t n, v, d;
-
-    if (!(n = strlen(opt_arg)))
-        invalid_opt_arg(opt_name, opt_arg);
-    v = parse_num(p = opt_arg, &q, 10);
-    d = PTR_OFFSET(q, p, n);
-    if (errno || (d != n))
-        invalid_opt_arg(opt_name, opt_arg);
-    if ((min > 0 && v < min) ||
-        (max > 0 && v > max))
-        illegal_opt_arg(opt_name, opt_arg);
-    return v;
-}
-
-static size_t parse_su_size_optarg(
-    const char* opt_name, const char* opt_arg,
-    size_t min, size_t max)
-{
-    const char *p, *q = NULL;
-    size_t n, v, d;
-
-    if (!(n = strlen(opt_arg)))
-        invalid_opt_arg(opt_name, opt_arg);
-    v = parse_num(p = opt_arg, &q, 10);
-    d = PTR_OFFSET(q, p, n);
-    if (errno ||
-        (d == 0) ||
-        (d < n - 1) ||
-        (d == n - 1 && *q != 'k' && *q != 'K' &&
-            *q != 'm' && *q != 'M'))
-        invalid_opt_arg(opt_name, opt_arg);
-    switch (*q) {
-    case '\0':
-        break;
-    case 'k':
-    case 'K':
-        STATIC(KB(1) <= SIZE_MAX);
-        if (!SIZE_MUL_NO_OVERFLOW(v, KB(1)))
-            illegal_opt_arg(opt_name, opt_arg);
-        v *= KB(1);
-        break;
-    case 'm':
-    case 'M':
-        STATIC(MB(1) <= SIZE_MAX);
-        if (!SIZE_MUL_NO_OVERFLOW(v, MB(1)))
-            illegal_opt_arg(opt_name, opt_arg);
-        v *= MB(1);
-        break;
-    }
-    if ((min > 0 && v < min) ||
-        (max > 0 && v > max))
-        illegal_opt_arg(opt_name, opt_arg);
-    return v;
-}
+#define SU_SIZE_OPT_NEED_PARSE_SIZE
+#define SU_SIZE_OPT_NEED_PARSE_SIZE_SU
+#define SU_SIZE_OPT_NEED_STATIC_FUNCS
+#include "lib/su-size-opt-impl.h"
 
 #ifdef DEBUG
 static size_t parse_debug_optarg(
@@ -1361,46 +1276,7 @@ static size_t parse_debug_optarg(
 
 #ifdef CONFIG_GETOPT_LONG2
 #include "long-opts-impl.h"
-
-static int getopt_long2(
-    int argc, char* const argv[],
-    const char* shorts, const struct option* longs,
-    int* index)
-{
-    const size_t n = sizeof(struct option);
-
-    static struct option l[2];
-    char *p, *q;
-    int k;
-
-    ASSERT(argc > 0);
-    ASSERT(argv != NULL);
-    ASSERT(shorts != NULL);
-    ASSERT(longs != NULL);
-
-    if (optind < argc) {
-        ASSERT(optind > 0);
-        p = argv[optind];
-
-        if (*p ++ == '-' && *p ++ == '-' && *p) {
-            if ((q = strchr(p, '=')))
-                *q = 0;
-
-            if ((k = lookup_long_opt(p)) >= 0) {
-                memcpy(l, &longs[k], n);
-                longs = l;
-            }
-            else
-                longs = l + 1;
-
-            if (q) *q = '=';
-        }
-    }
-
-    return getopt_long(argc, argv, shorts, longs, index);
-}
-
-#define getopt_long getopt_long2
+#include "lib/getopt2-impl.h"
 #endif // CONFIG_GETOPT_LONG2
 
 #if defined(DEBUG) && defined(DEBUG_GETOPT_LONG)
@@ -1409,7 +1285,8 @@ static void print_getopt_long_debug(int opt)
 {
     fprintf(stderr, "optind=%d optarg=", optind);
     if (optarg)
-        pretty_print_string(stderr, (const uchar_t*) optarg,
+        pretty_print_string(stderr,
+            PTR_UCHAR_CAST_CONST(optarg),
             strlen(optarg), pretty_print_string_quotes);
     else
         fputs("(nil)", stderr);
@@ -1441,7 +1318,9 @@ const struct options_t* options(int argc, char* argv[])
         .type_lib         = NULL,
         .type_name        = NULL,
         .sigpipe_act      = options_sigpipe_act_error,
+        .filter_libs      = false,
         .empty_input      = false,
+        .raw_strings      = false,
         .liter_value      = false,
         .surrogate_pairs  = false,
         .valid_utf8       = false,
@@ -1452,6 +1331,7 @@ const struct options_t* options(int argc, char* argv[])
         .quote_text       = false,
         .trim_spaces      = false,
         .newline_sep      = false,
+        .no_warning       = false,
         .no_error         = false,
         .verbose          = false,
 #undef  CASE
@@ -1476,7 +1356,7 @@ const struct options_t* options(int argc, char* argv[])
         // stev: actions:
         quiet_parser_act     = 'O',
         echo_parser_act      = 'E',
-        from_ast_printer_act = 'F',
+        from_ast_printer_act = 'M',
         pretty_printer_act   = 'P',
         terse_printer_act    = 'R',
         type_printer_act     = 'Y',
@@ -1485,14 +1365,17 @@ const struct options_t* options(int argc, char* argv[])
         json2_act            = 'J',
         //!!!2JON two_json_act = '2',
         typelib_func_act     = 'T',
+        filterlib_exec_act   = 'F',
 
         // stev: options:
         type_def_opt         = 'd',
         type_lib_opt         = 't',
-        type_name_opt        = 'N',
+        type_name_opt        = 'n',
         input_file_opt       = 'i',
         sigpipe_act_opt      = 'p',
+        filter_libs_opt      = 'f',
         empty_input_opt      = 'y',
+        raw_strings_opt      = 'w',
         liter_value_opt      = 'l',
         surrogate_pairs_opt  = 's',
         valid_utf8_opt       = 'u',
@@ -1500,7 +1383,7 @@ const struct options_t* options(int argc, char* argv[])
         ascii_only_opt       = 'a',
         multi_obj_opt        = 'm',
         print_dots_opt       = 'o',
-        quote_text_opt       = 'Q',
+        quote_text_opt       = 'q',
         trim_spaces_opt      = 'r',
         version_opt          = 'v',
         verbose_opt          = 'V',
@@ -1512,6 +1395,7 @@ const struct options_t* options(int argc, char* argv[])
 #include "opt-sizes.def"
 
         newline_sep_opt      = 128,
+        no_warning_opt,
         no_error_opt,
         dump_opts_opt,
 
@@ -1521,7 +1405,9 @@ const struct options_t* options(int argc, char* argv[])
 #include "opt-sizes.def"
 
         no_type_check_opt,
+        no_filter_libs_opt,
         no_empty_input_opt,
+        no_raw_strings_opt,
         no_liter_value_opt,
         no_surrogate_pairs_opt,
         no_valid_utf8_opt,
@@ -1532,6 +1418,7 @@ const struct options_t* options(int argc, char* argv[])
         no_quote_text_opt,
         no_trim_spaces_opt,
         no_newline_sep_opt,
+        no_no_warning_opt,
         no_no_error_opt,
         no_verbose_opt,
 
@@ -1564,6 +1451,7 @@ const struct options_t* options(int argc, char* argv[])
         { "type-print",             0,       0, type_printer_act },
         { "ast-print",              0,       0, ast_printer_act },
         { "typelib-func",           2,       0, typelib_func_act },
+        { "filterlib-exec",         0,       0, filterlib_exec_act },
         { "escape-json",            0,       0, escape_json_act },
         { "json2",                  0,       0, json2_act },
 //!!!2JSON		{ "2json",                  0,       0, two_json_act },
@@ -1572,10 +1460,14 @@ const struct options_t* options(int argc, char* argv[])
         { "type-lib",               1,       0, type_lib_opt },
         { "type-name",              2,       0, type_name_opt },
         { "no-type-check",          0,       0, no_type_check_opt },
+        { "filter-libs",            0,       0, filter_libs_opt },
+        { "no-filter-libs",         0,       0, no_filter_libs_opt },
         { "input-file",             1,       0, input_file_opt },
         { "sigpipe-action",         1,       0, sigpipe_act_opt },
         { "empty-input",            0,       0, empty_input_opt },
         { "no-empty-input",         0,       0, no_empty_input_opt },
+        { "raw-strings",            0,       0, raw_strings_opt },
+        { "no-raw-strings",         0,       0, no_raw_strings_opt },
         { "liter-value",            0,       0, liter_value_opt },
         { "no-liter-value",         0,       0, no_liter_value_opt },
         { "literal-value",          0,       0, liter_value_opt },
@@ -1604,6 +1496,8 @@ const struct options_t* options(int argc, char* argv[])
         { "no-newline-sep",         0,       0, no_newline_sep_opt },
         { "dump-options",           0,       0, dump_opts_opt },
         { "version",                0,       0, version_opt },
+        { "warning",                0,       0, no_no_warning_opt },
+        { "no-warning",             0,       0, no_warning_opt },
         { "error",                  0,       0, no_no_error_opt },
         { "no-error",               0,       0, no_error_opt },
         { "verbose",                0,       0, verbose_opt },
@@ -1630,26 +1524,33 @@ const struct options_t* options(int argc, char* argv[])
     };
     static const char short_opts[] =
         ":"
-        "AEFJOPRST::Y"
+        "AEFJMOPRST::Y"
 #undef  CASE
 #define CASE(type, id, name, pad, letter, desc, help, min, max, def) \
         ARG_LETTER(letter)
 #include "opt-sizes.def"
-        "acd:i:lmN::oprQst:uvVy";
+        "acd:fi:lmn::opqrst:uvVwy";
 
     struct bits_opts_t {
         bits_t dump: 1;
         bits_t usage: 1;
         bits_t usage_sizes: 1;
         bits_t version: 1;
+        bits_t has_filters: 1;
     };
     struct bits_opts_t bits = {
         .dump        = false,
         .usage       = false,
         .usage_sizes = false,
         .version     = false,
+        .has_filters = false,
     };
 
+    struct options_filter_t fopt = {
+        .argc = argc,
+        .argv = argv
+    };
+    struct options_filter_t popt;
     int opt;
 
 #define argv_optind()                      \
@@ -1666,6 +1567,12 @@ const struct options_t* options(int argc, char* argv[])
         (char) optopt;                  \
     })
 
+    if (update_filter_opt(&fopt, &popt)) {
+        bits.has_filters = true;
+        ASSERT(popt.argv == argv);
+        argc = popt.argc;
+    }
+
     opterr = 0;
     optind = 1;
     while ((opt = getopt_long(argc, argv, short_opts,
@@ -1693,7 +1600,14 @@ const struct options_t* options(int argc, char* argv[])
                     &opts.type_name);
             break;
         case type_name_opt:
-            opts.type_name = optarg;
+            opts.type_name =
+                optarg && *optarg ? optarg : NULL;
+            break;
+        case filter_libs_opt:
+            opts.filter_libs = true;
+            break;
+        case no_filter_libs_opt:
+            opts.filter_libs = false;
             break;
         case sigpipe_act_opt:
             opts.sigpipe_act = parse_sigpipe_act_optarg(
@@ -1705,6 +1619,12 @@ const struct options_t* options(int argc, char* argv[])
         case no_empty_input_opt:
             opts.empty_input = false;
             break;
+        case raw_strings_opt:
+            opts.raw_strings = true;
+            break;
+        case no_raw_strings_opt:
+            opts.raw_strings = false;
+            break;
         case liter_value_opt:
             opts.liter_value = true;
             break;
@@ -1776,7 +1696,7 @@ const struct options_t* options(int argc, char* argv[])
 #include "opt-sizes.def"
 
 #ifdef DEBUG
-#define CASE_DEBUG_(t, n)                              \
+#define CASE_DEBUG__(t, n)                             \
         case debug_ ## t ## _opt: {                    \
                 size_t __n;                            \
                 __n = parse_debug_optarg(n, optarg);   \
@@ -1784,6 +1704,8 @@ const struct options_t* options(int argc, char* argv[])
                 opts.debug_ ## t = __n;                \
             break;                                     \
         }
+#define CASE_DEBUG_(t, n) CASE_DEBUG__(t, "debug-" n)
+
 #define CASE_DEBUG(t) CASE_DEBUG_(t, #t)
 #define CASE_DEBUG2   CASE_DEBUG_
 
@@ -1813,6 +1735,12 @@ const struct options_t* options(int argc, char* argv[])
         case version_opt:
             bits.version = true;
             break;
+        case no_warning_opt:
+            opts.no_warning = true;
+            break;
+        case no_no_warning_opt:
+            opts.no_warning = false;
+            break;
         case no_error_opt:
             opts.no_error = true;
             break;
@@ -1861,8 +1789,11 @@ const struct options_t* options(int argc, char* argv[])
 
         case typelib_func_act:
             opts.object = OBJ_JSON_TYPELIB;
-            opts.typelib_func = 
-                parse_typelib_func_optarg("typelib-func", optarg);
+            opts.typelib_func = parse_typelib_func_optarg(
+                "typelib-func", optarg);
+            break;
+        case filterlib_exec_act:
+            opts.object = OBJ_JSON_FILTERLIB;
             break;
         case help_sizes_opt:
             bits.usage_sizes = true;
@@ -1899,22 +1830,31 @@ const struct options_t* options(int argc, char* argv[])
         }
     }
 
-    ASSERT(argc >= optind);
+    ASSERT(optind > 0);
+    ASSERT(optind <= argc);
 
     argc -= optind;
     argv += optind;
 
     if (argc > 0) {
-        opts.input = *argv;
+        opts.input = *argv ++;
         argc --;
-        argv ++;
     }
 
-    opts.argc = argc;
+    opts.argc = INT_AS_SIZE(argc);
     opts.argv = argv;
 
-    if (opts.input && !strcmp(opts.input, "-"))
+    if (bits.has_filters &&
+        (opts.filter_libs ||
+         opts.object == OBJ_JSON_FILTERLIB))
+        parse_filters(fopt, opts.filters, &opts.n_filters,
+            ARRAY_SIZE(opts.filters));
+
+    if (opts.input && !strcmp(opts.input, "-")) {
+        if (opts.object == OBJ_JSON_FILTERLIB)
+            error("input filter library cannot be stdin");
         opts.input = NULL;
+    }
 
     if (bits.version)
         version();
@@ -1943,10 +1883,65 @@ const struct options_t* options(int argc, char* argv[])
             opts.sizes_error_ctxt_size,
             opts.sizes_error_buf_max);
 
-    if (opts.object == OBJ_JSON_TYPELIB &&
-        opts.multi_obj)
-        warning("`-m|--multi-object' is ignored when "
-            "given along with `-T|--typelib-func'");
+    if (opts.object == OBJ_JSON_TYPELIB) {
+        if (!opts.no_warning && opts.multi_obj)
+            warning("`-m|--multi-object' is ignored when "
+                "given along with `-T|--typelib-func'");
+    }
+
+    if (opts.object == OBJ_JSON_FILTERLIB) {
+        if (!opts.no_warning &&
+             opts.n_filters > 1 && opts.input == NULL)
+            warning("action `-F|--filterlib-exec': "
+                "multiple filter command line "
+                "arguments groups given");
+
+        if (!opts.no_warning &&
+             opts.n_filters > 0 && opts.input != NULL)
+            warning("action `-F|--filterlib-exec': "
+                "ignoring the filter command line arguments "
+                "since input file was given");
+
+        if (opts.n_filters == 0 && opts.input == NULL)
+            error("action `-F|--filterlib-exec' requires "
+                "at least one filter command line "
+                "arguments group");
+
+        if (opts.input != NULL) {
+            static char* argv[3] = { [1] = "--help" };
+
+            argv[0] = CONST_CAST(opts.input, char);
+            opts.filters->argv = argv;
+            opts.filters->argc = 2;
+            opts.n_filters = 1;
+        }
+    }
+
+    if (!opts.no_warning &&
+         bits.has_filters &&
+        !opts.filter_libs &&
+         opts.object != OBJ_JSON_FILTERLIB)
+        warning("ignoring the filter command line arguments "
+            "since neither `-F|--filterlib-exec' nor "
+            "`-f|--filter-libs' were given");
+
+    if (!opts.no_warning &&
+         bits.has_filters &&
+         opts.filter_libs &&
+         opts.object != OBJ_JSON_PARSER &&
+         opts.object != OBJ_JSON_PRINTER &&
+         opts.object != OBJ_JSON2)
+        warning("ignoring the filter command line arguments "
+            "since none of JSON parser, printer or flattener "
+            "options were given");
+
+    if (opts.filter_libs &&
+        opts.n_filters == 0 &&
+        opts.object != OBJ_JSON_ESCAPE &&
+        opts.object != OBJ_JSON_TYPELIB &&
+        opts.object != OBJ_JSON_FILTERLIB)
+        error("option `-f|--filter-libs' requires at least "
+            "one filter command line arguments group");
 
     init_sigpipe_action(opts.sigpipe_act);
 
