@@ -1,4 +1,4 @@
-// Copyright (C) 2016, 2017, 2018  Stefan Vargyas
+// Copyright (C) 2016, 2017, 2018, 2019  Stefan Vargyas
 // 
 // This file is part of Json-Type.
 // 
@@ -34,14 +34,19 @@ struct obj_json_printer_t
     bits_t first_nl: 1;
     bits_t printed_out: 1;
     bits_t skip_indent: 1;
+    bits_t raw_strings: 1;
     bits_t print_dots: 1;
     bits_t newline_sep: 1;
 #ifdef DEBUG
     bits_t debug: debug_bits;
 #endif
     size_t print_flags;
-
     size_t indent_level;
+
+    const uchar_t* ast_delim;
+
+    const uchar_t* (*get_delim)(
+        struct obj_json_printer_t*);
 };
 
 #endif/*__JSON_PRINTER_H*/
