@@ -223,7 +223,7 @@ static bool json2_value_sep(struct obj_json2_t* this)
     return true;
 }
 
-static struct json_handler_t handler = {
+static const struct json_handler_t handler = {
     .null_func         = (json_null_func_t)         json2_null,
     .boolean_func      = (json_boolean_func_t)      json2_bool,
     .number_func       = (json_number_func_t)       json2_number,
